# Code based on...
# https://kidscancode.org/godot_recipes/2d/car_steering/

tool
extends KinematicBody2D
class_name Vehicle

export var camera_enabled : bool = false	setget _set_camera_enabled
export var forward_axel : float = 10 		setget _set_forward_axel
export var rear_axel : float = 10 			setget _set_rear_axel
export (float, 0.1, 360.0, 0.1) var max_steering_angle = 15 setget _set_max_steering_angle
export var max_engine_power : float = 800 	setget _set_max_engine_power
export (float, 0.0, 1.0, 0.001) var reverse_engine_multiplier = 0.85
export var max_breaking_power : float = 450	setget _set_max_breaking_power
export (float, 0.01, 1.0, 0.01) var friction = 0.9
export (float, 0.001, 1.0, 0.0005) var drag = 0.0015

var velocity = Vector2.ZERO
var acceleration = Vector2.ZERO

var steering_angle = 0.0
var engine_power = 0.0
var breaking_power = 0.0
var reverse = false

onready var camera = $Camera2D

# -----------------------------------------------------------------------------
# Setter/Getter Methods
# -----------------------------------------------------------------------------

func _set_camera_enabled(e : bool) -> void:
	camera_enabled = e
	if camera:
		camera.current = e

func _set_forward_axel(a : float) -> void:
	if a > 0:
		forward_axel = a

func _set_rear_axel(a : float) -> void:
	if a > 0:
		rear_axel = a

func _set_max_steering_angle(a : float) -> void:
	max_steering_angle = a
	if steering_angle > a:
		steering_angle = a
	elif steering_angle < -a:
		steering_angle = -a

func _set_max_engine_power(p : float) -> void:
	if p > 0.0:
		max_engine_power = p
		if engine_power > p:
			engine_power = p

func _set_max_breaking_power(p : float) -> void:
	if p > 0.0:
		max_breaking_power = p
		if breaking_power > p:
			breaking_power = p

# -----------------------------------------------------------------------------
# Godot Override Methods
# -----------------------------------------------------------------------------

func _ready():
	camera.current = camera_enabled
	if Engine.editor_hint:
		$Sprite.show_behind_parent = true
		$Collision.visible = false

func _draw():
	if Engine.editor_hint:
		var wheel_base_color = Color(1.0, 0.5, 0.0)
		draw_line(
			Vector2(0, -forward_axel),
			Vector2(0, rear_axel),
			wheel_base_color,
			2
		)

func _process(delta):
	update()

func _physics_process(delta):
	if Engine.editor_hint:
		return

	# TODO: Reverse engine_power sign when toggling reverse!
	acceleration = -transform.y * engine_power
	
	_apply_friction()
	print(acceleration)
	_calculate_heading(delta)
	
	if reverse:
		velocity += -acceleration * delta
	else:
		velocity += acceleration * delta
	velocity = move_and_slide(velocity)

# -----------------------------------------------------------------------------
# Private "Helper" Methods
# -----------------------------------------------------------------------------

func _calculate_heading(delta : float) -> void:
	var rear_wheel = position + transform.y * rear_axel
	var fore_wheel = position - transform.y * forward_axel
	rear_wheel += velocity * delta
	fore_wheel += velocity.rotated(deg2rad(steering_angle)) * delta
	var heading = (fore_wheel - rear_wheel).normalized()
	var d = heading.dot(velocity.normalized())
	if d > 0:
		velocity = heading * velocity.length()
	if d < 0:
		velocity = -heading * velocity.length()
	#velocity = heading * velocity.length()
	rotation = (heading.angle() + deg2rad(90))

func _apply_friction() -> void:
	var speed = velocity.length()
	if velocity.length() < 5:
		velocity = Vector2.ZERO
		speed = 0
	var force_friction = (-friction) * velocity
	var force_drag = (-drag) * speed * velocity
	acceleration += force_friction + force_drag

# -----------------------------------------------------------------------------
# Public Methods
# -----------------------------------------------------------------------------

func get_wheel_base() -> float:
	return forward_axel + rear_axel

func get_steering_angle() -> float:
	return steering_angle

func set_steering(v : float) -> void:
	v = (clamp(v, -1.0, 1.0) + 1) / 2.0
	steering_angle = (max_steering_angle * 2 * v) - max_steering_angle

func is_reverse() -> bool:
	return reverse

func set_reverse(r : bool) -> void:
	reverse = r

func toggle_reverse() -> bool:
	reverse = not reverse
	return reverse

func get_engine_power() -> float:
	return engine_power

func set_throttle(v : float) -> void:
	v = clamp(v, 0.0, 1.0)
	var epm = 1.0
	if reverse:
		epm = reverse_engine_multiplier
	engine_power = max_engine_power * epm * v
	
	
# -----------------------------------------------------------------------------
# Event Handlers
# -----------------------------------------------------------------------------


